/**
 * Custom sync proxy
 */
Ext.data.FHStorageProxy = Ext.extend(Ext.data.Proxy, {

    /**
     * @cfg id Unique id
     */
    id: undefined,

    constructor: function () {

	   Ext.data.FHStorageProxy.superclass.constructor.apply(this, arguments);


        this.addEvents(
        'loadexception', 'readexception', 'createexception'

        );

        // if an id is not given, try to use the store's id instead
        this.id = this.id || (this.store ? this.store.storeId : undefined);

        if (this.id == undefined) {
            throw "No unique id was provided to the local storage proxy.";
        }


    },

    /**
     * Create records
     */
    create: function (operation, callback, scope) {

        $fh.log({
            message: 'ProxyId: ' + this.id + ',operation: ' + operation.action + ',operation source: ' + operation.source + ',operation records: ' + (operation.records ? operation.records.length : 0)
        });

        var me = this;

        switch (operation.source) {
        case "local":
            return this.doSaveLocal(operation, callback, scope);
            break;
        case "remote":
            return this.doSaveRemote(operation, callback, scope);
            break;
        case "localRemote":
            return this.doSaveLocal(operation, callback, scope, function () {
                me.doSaveRemote(operation, callback, scope);
            });
            break;
        case "remoteLocal":
            return this.doSaveRemote(operation, callback, scope, function () {
                me.doSaveLocal(operation, callback, scope);
            });
            break;
        default:
            return this.doSaveLocal(operation, callback, scope);
            break;
        };


    },

    /**
     * Load records
     */    
    read: function (operation, callback, scope) {

        var me = this;

        switch (operation.source) {
        case "local":
            return this.doLoadLocal(operation, callback, scope);
            break;
        case "remote":
            return this.doLoadRemote(operation, callback, scope);
            break;
        case "localRemote":
            return this.doLoadLocal(operation, callback, scope, function () {
                me.doLoadRemote(operation, callback, scope);
            });
            break;
        case "remoteLocal":
            return this.doLoadRemote(operation, callback, scope, function () {
                me.doLoadLocal(operation, callback, scope);
            });
            break;
        default:
            return this.doLoadLocal(operation, callback, scope);
            break;
        };

    },
    
    /**
     * Update records
     */    

    update: function () {
        return this.create.apply(this, arguments);
    },
    
    /**
     * Destroy records
     */    

    destroy: function (operation, callback, scope) {

        $fh.log({
            message: 'ProxyId: ' + this.id + ',operation: ' + operation.action + ',operation source: ' + operation.source + ',operation records: ' + (operation.records ? operation.records.length : 0)
        });

        var me = this;

        switch (operation.source) {
        case "local":
            return this.doRemoveLocal(operation, callback, scope);
            break;
        case "remote":
            return this.doRemoveRemote(operation, callback, scope);
            break;
        case "localRemote":
            return this.doRemoveLocal(operation, callback, scope, function () {
                me.doRemoveRemote(operation, callback, scope);
            });
            break;
        case "remoteLocal":
            return this.doRemoveRemote(operation, callback, scope, function () {
                me.doRemoveLocal(operation, callback, scope);
            });
            break;
        default:
            return this.doRemoveLocal(operation, callback, scope);
            break;
        };

    },

    /**
     * Prepares callback
     * @param {Ext.data.Operation}
     *            operation The Operation being executed
     * @param {Function}
     *            callback The callback function to be called when the request
     *            completes. This is usually the callback passed to doRequest
     * @param {Object}
     *            scope The scope in which to execute the callback function
     * @return {Function} The callback function if source is 'remoteLocal' or 'localRemote'
     */
    createRequestCallback: function (operation, callback, scope, syncCallback) {


        var me = this;

        return function (response) {

            var result = {};
            var reader = me.getReader();
            var resultWrapper = {};

            var hashRes;

            //prepare the text for json reader (must use json reader to activate associations)
            //if response from $fh.data
            if (response && response != null && response.val && response.val != null && response.val != "") {
                resultWrapper.responseText = response.val;
                try {
                    result = reader.read(resultWrapper);
                } catch (e) {}
                var localResponse = JSON.parse(response.val);
                //update hash
                hashRes = localResponse.hash;
            }
            //if response comes from remote source
            else if (response && response != null && !response.val) {
                resultWrapper.responseText = JSON.stringify(response);
                try {
                    result = reader.read(resultWrapper);
                } catch (e) {}
                hashRes = response.hash;
            }

            (scope || me).hash = hashRes || (scope || me).hash;

            //default
            records = (operation.resultSet) ? operation.resultSet.records : operation.records || scope.data.items;

            if (result && result.records) {
                records = result.records;
            }

            // if action update - records are going to be appended else existing data will be cleared 
            operation.addRecords = (operation.action == 'update') ? true : false;


            operation.addRecords = (operation.append) ? true : false;


            operation.resultSet = new Ext.data.ResultSet({
                records: records,
                total: records.length,
                loaded: true
            });

            operation.setSuccessful();
            operation.setCompleted();


            //update encrypted response of the store
            //(scope || me).encryptedRecords = response.encryptedRecords || (scope || me).encryptedRecords || "";
            // this callback is the one that was passed to the 'read' or 'write'
            if (typeof callback == 'function' && !(typeof syncCallback == 'function' && operation.defferCallback)) {
                callback.call(scope || me, operation);
                //bug fix - update store snapshot
                (scope || me).snapshot = (scope || me).data.clone();
            }

            //this will call other storage if operation source 'localRemote' or 'remoteLocal'        
            //it must fire only if the current store records has expired
            if (typeof syncCallback == 'function') {

                //var nowInSec = new Date().getTime();
                //var timestampInSec = timestamp || 0;
                //var expiryInSec = (scope || me).expiry || 0;

                //if (nowInSec - timestampInSec > expiryInSec ){ 
                syncCallback.call(operation, callback, scope);
                //}
            }

            return me;

        };

    },
    /**
     * Prepares errback
     * @param {Ext.data.Operation}
     *            operation The Operation being executed
     * @param {Function}
     *            callback The callback function to be called when the request
     *            completes. This is usually the callback passed to doRequest
     * @param {Object}
     *            scope The scope in which to execute the callback function
     * @return {Function} The callback function if source is 'remoteLocal' or 'localRemote'
     */
    createRequestErrback: function (operation, callback, scope, syncCallback) {

        var me = this;

        return function (response) {

            (scope || me).fireEvent(operation.action + 'exception', this, response, operation);
   
            var message = (!response || null==response) ? "" :  response.message || "";
            
            operation.setException(message);
            operation.setCompleted();

            // fix - now existing data won't be cleared - addRecords is like append [] 
            operation.addRecords = true;
            operation.resultSet = new Ext.data.ResultSet({
                records: [],
                total: 0,
                loaded: false
            });

            // this callback is the one that was passed to the 'read' or 'write'
            if (typeof callback == 'function' && !(typeof syncCallback == 'function' && operation.defferCallback)) {
                callback.call(scope || me, operation);
            }

            //this will call other storage if operation source 'localRemote' or 'remoteLocal'        
            //it must fire only if the current store records has expired
            if (typeof syncCallback == 'function') {

                //var nowInSec = new Date().getTime();
                //var timestampInSec = timestamp || 0;
                //var expiryInSec = (scope || me).expiry || 0;

                //if (nowInSec - timestampInSec > expiryInSec ){ 
                syncCallback.call(operation, callback, scope);
                //}
            }

            return me;
        };
    },
    

    /**
     * Encodes the array of {@link Ext.util.Sorter} objects into a string to be sent in the request url. By default, 
     * this simply JSON-encodes the sorter data
     * @param {Array} sorters The array of {@link Ext.util.Sorter Sorter} objects
     * @return {String} The encoded sorters
     */
    encodeSorters: function (sorters) {
        var min = [],
            length = sorters.length,
            i;

        for (i = 0; i < length; i++) {
            min[i] = {
                property: sorters.items[i].property,
                direction: sorters.items[i].direction
            };
        }

        return min;
    },


    /**
     * Encodes the array of {@link Ext.util.Filter} objects into a string to be sent in the request url. By default, 
     * this simply JSON-encodes the filter data
     * @param {Array} store The store to read the filters from
     * @param {Boolean} excluded_key The property of filter to exclude 
     * @param {Array} min Object that holds the extracted filters
     * @return {String} The encoded filters
     */
    encodeFilters: function (store, excluded_key, min) {
        var min = min || [],
            length = store.filters.length,
            i;

        for (i = 0; i < length; i++) {
            var property = store.filters.items[i].property;
            var value = store.filters.items[i].value;
            if (property != excluded_key) {
                min.push({
                    property: property,
                    value: value.source || value
                });
            }

        }
        //remote filters
        length = store.remoteFilters ? store.remoteFilters.length : 0;
        for (i = 0; i < length; i++) {
            var property = store.remoteFilters[i].property;
            var value = store.remoteFilters[i].value;
            if (property != excluded_key) {
                min.push({
                    property: property,
                    value: value.source || value
                });
            }

        }

/*
      var associations = store.model.prototype.associations || {};
      if (associations.items && associations.items.length){
          for(var a=0;a<associations.items.length;a++){
              var item = associations.items[a];
              if(item.type == "hasMany"){
                var storeName = item.storeName;
                var substore = store.data.items.()[associationName]();
                      this.encodeFilters(substore, item.foreignKey, min);              
              }
          }
      }   
      */
        return min;
    },

    /**
     * Proxy {@link #save} with local source pass through to FH.LocalStorage.
     * 
     * @param {Ext.data.Operation}
     *            operation The Ext.data.Operation object
     * @param {Function}
     *            callback The callback function to call when the Operation has
     *            completed
     * @param {Object}
     *            scope The scope in which to execute the callback
     * @param {Function}
     *            syncCallback The callback function to call when data can have two sources           
     */
    doSaveLocal: function (operation, callback, scope, syncCallback) {

      var result = {};
      var writer = this.getWriter();
        var val = {
            'hash': (scope || this).hash
        };

        var records = writer.write(operation.records || (scope || me).data.items || []);
        val.records = records;

        FHLocalStorage.save(operation.act, val, this.createRequestCallback(operation, callback, scope, syncCallback), this.createRequestErrback(operation, callback, scope, syncCallback));
        return this;

    },
    
    /**
     * Proxy {@link #save} with remote source pass through to FHServerStorage.
     * 
     * @param {Ext.data.Operation}
     *            operation The Ext.data.Operation object
     * @param {Function}
     *            callback The callback function to call when the Operation has
     *            completed
     * @param {Object}
     *            scope The scope in which to execute the callback
     * @param {Function}
     *            syncCallback The callback function to call when data can have two sources           
     */    

    doSaveRemote: function (operation, callback, scope, syncCallback) {
      

      var writer = this.getWriter();
        var val = {
            'hash': (scope || this).hash,
            'act' : operation.act
        };

        var records = writer.write(operation.records || (scope || me).data.items || []);
        val.records = records;


        FHServerStorage.save(val, this.createRequestCallback(operation, callback, scope, syncCallback), this.createRequestErrback(operation, callback, scope, syncCallback));


        return this;

    },
    
    /**
     * Proxy {@link #read} and {@link #update} with local source pass through to FH.LocalStorage.
     * 
     * @param {Ext.data.Operation}
     *            operation The Ext.data.Operation object
     * @param {Function}
     *            callback The callback function to call when the Operation has
     *            completed
     * @param {Object}
     *            scope The scope in which to execute the callback
     * @param {Function}
     *            syncCallback The callback function to call when data can have two sources           
     */    

    doLoadLocal: function (operation, callback, scope, syncCallback) {

        FHLocalStorage.load(operation.act, this.createRequestCallback(operation, callback, scope, syncCallback), this.createRequestErrback(operation, callback, scope, syncCallback));

    },
    
    /**
     * Proxy {@link #read} and {@link #update} with remote source pass through to FHServerStorage.
     * 
     * @param {Ext.data.Operation}
     *            operation The Ext.data.Operation object
     * @param {Function}
     *            callback The callback function to call when the Operation has
     *            completed
     * @param {Object}
     *            scope The scope in which to execute the callback
     * @param {Function}
     *            syncCallback The callback function to call when data can have two sources           
     */    

    doLoadRemote: function (operation, callback, scope, syncCallback) {

        var val = {
            'act': operation.act,
            'sorters': this.encodeSorters(scope.sorters),
            'filters': this.encodeFilters(scope),
            'hash' : (scope|| me).hash || "",
            'debug': true
        };
        if(operation.action == 'update'){
          //VERY IMPORTANT: We should use writer here, but I use update only to check if excludedProperty has changed, so
          //here I clone the record.data - without associations - and exclude the excludedProperty, to save bandwidth
          //no time to make it GENERIC, don't use update then if you have to exclude more than one excludedProperty or use associations (or change the code :) )
          var records = [];
          var currentRecords = (operation.resultSet) ? operation.resultSet.records : operation.records || scope.data.items;
          currentRecords = !currentRecords.length ? scope.data.items : currentRecords;
            for (var r=0; r < currentRecords.length; r++) {
              var obj = currentRecords[r].data;
              var temp = obj.constructor(); 
              for(var key in obj){
                var value = "";
                if(!scope.excludedProperty || key != scope.excludedProperty){
                  value = obj[key];
                }
                  temp[key] = value;
              }
              records.push(temp);
            }          
            val.records = records;
        }
        FHServerStorage.load(val, this.createRequestCallback(operation, callback, scope, syncCallback), this.createRequestErrback(operation, callback, scope, syncCallback));

    },
    
    /**
     * Proxy {@link #destroy} with local source pass through to FH.LocalStorage.
     * 
     * @param {Ext.data.Operation}
     *            operation The Ext.data.Operation object
     * @param {Function}
     *            callback The callback function to call when the Operation has
     *            completed
     * @param {Object}
     *            scope The scope in which to execute the callback
     * @param {Function}
     *            syncCallback The callback function to call when data can have two sources           
     */    

    doRemoveLocal: function (operation, callback, scope, syncCallback) {

        FHLocalStorage.load(operation.act, this.createRequestCallback(operation, callback, scope, syncCallback), this.createRequestErrback(operation, callback, scope, syncCallback));

    },
    
    /**
     * Proxy {@link #destroy} with remote source pass through to FHServerStorage.
     * 
     * @param {Ext.data.Operation}
     *            operation The Ext.data.Operation object
     * @param {Function}
     *            callback The callback function to call when the Operation has
     *            completed
     * @param {Object}
     *            scope The scope in which to execute the callback
     * @param {Function}
     *            syncCallback The callback function to call when data can have two sources           
     */    

    doRemoveRemote: function (operation, callback, scope, syncCallback) {

        FHServerStorage.load(operation.act, this.createRequestCallback(operation, callback, scope, syncCallback), this.createRequestErrback(operation, callback, scope, syncCallback));

    }

});


Ext.data.ProxyMgr.registerType('fhsync', Ext.data.FHStorageProxy);
