

  /**
   * Creates an object that executes a function and fires events depending if the function callback or errback have been invoked.
   */
Ext.ux.Condition = Ext.extend(Ext.util.Observable, {

  constructor: function (config) {

    Ext.ux.Condition.superclass.constructor.call(this, config);
    /**
     * @event conditionTrue Triggered after {@link Ext.ux.Condition#fhFunction} invokes callback.
     * @param res Argument the callback function gets.
     */
    /**
     * @event conditionFalse Triggered after {@link Ext.ux.Condition#fhFunction} invokes errback.
     * @param res Argument the errback function gets.
     */
    this.addEvents('conditionTrue', 'conditionFalse');

  },
  
  
  /**
   * @cfg {String} id Required property. Unique identifier of the condition if used as one of many conditions of the {@link Ext.ux.AsyncActionQueue} queue.
   */
  id: undefined,
  
  /**
   * @cfg {Function} fhFunction Required property. Must reference a function that accepts three argumants: params, callback and errback.
   */
  fhFunction: function () {},
  
  /**
   * @cfg {Object} fhFunctionObject Optional property. It holds the params object that will be pass as the {@link Ext.ux.Condition#fhFunction} first argument.
   */  
  fhFunctionObject: {},

  /**
   * @cfg {Function} trueCallback Optional property. It holds the callback function that will executed if the {@link Ext.ux.Condition#fhFunction} invokes callback.
   */  
  trueCallback: undefined,
  /**
   * @cfg {Function} trueCallback Optional property. It holds the errback function that will be executed if the {@link Ext.ux.Condition#fhFunction} invokes errback.
   */   
  falseCallback: undefined,
  /**
   * @cfg {Object} trueMsg Optional property. If {@link Ext.ux.Condition#trueCallback} is not specified, this property may hold the config object for the Ext.MessageBox that should be displayed if the {@link Ext.ux.Condition#fhFunction} invokes callback.
   */
  trueMsg: undefined,
  /**
   * @cfg {Function} trueMsgCallback Optional property. This function will be called if the 'OK' button was pressed on the Ext.MessageBox with {@link Ext.ux.Condition#trueMsg} config.
   */  
  trueMsgCallback: undefined,
  /**
   * @cfg {Boolean} trueMsgCallbackStop Optional property. If set to false, none of the conditions in the {@link Ext.ux.AsyncActionQueue} queue will be called if the {@link Ext.ux.Condition#trueMsgCallback} function was invoked.
   */  
  trueMsgCallbackStop: false,
  /**
   * @cfg {Function} trueMsgErrback Optional property. This function will be called if the 'CANCEL' button was pressed on the Ext.MessageBox with {@link Ext.ux.Condition#trueMsg} config.
   */    
  trueMsgErrback: undefined,
  /**
   * @cfg {Boolean} trueMsgErrbackStop Optional property. If set to false, none of the conditions in the {@link Ext.ux.AsyncActionQueue} queue will be called if the {@link Ext.ux.Condition#trueMsgErrback} function was invoked.
   */   
  trueMsgErrbackStop: false,
  /**
   * @cfg {Object} falseMsg Optional property. If {@link Ext.ux.Condition#falseCallback} is not specified, this property may hold the config object for the Ext.MessageBox that should be displayed if the {@link Ext.ux.Condition#fhFunction} invokes errback.
   */
  falseMsg: undefined,
  /**
   * @cfg {Function} falseMsgCallback Optional property. This function will be called if the 'OK' button was pressed on the Ext.MessageBox with {@link Ext.ux.Condition#falseMsg} config.
   */   
  falseMsgCallback: undefined,
  /**
   * @cfg {Boolean} falseMsgCallbackStop Optional property. If set to false, none of the conditions in the {@link Ext.ux.AsyncActionQueue} queue will be called if the {@link Ext.ux.Condition#falseMsgCallback} function was invoked.
   */   
  falseMsgCallbackStop: false,
  /**
   * @cfg {Function} falseMsgErrback Optional property. This function will be called if the 'CANCEL' button was pressed on the Ext.MessageBox with {@link Ext.ux.Condition#falseMsg} config.
   */   
  falseMsgErrback: undefined,
  /**
   * @cfg {Boolean} falseMsgErrbackStop Optional property. If set to false, none of the conditions in the {@link Ext.ux.AsyncActionQueue} queue will be called if the {@link Ext.ux.Condition#falseMsgErrback} function was invoked.
   */   
  falseMsgErrbackStop: false,
  
  
  /**
   * Wrapper around function. Executes the {@link Ext.ux.Condition#fhFunction} and fires {@link Ext.ux.Condition#event-conditionTrue} or {@link Ext.ux.Condition#event-conditionFalse} event.
   */ 
  runCondition: function () {

    var me = this;

    if (typeof this.fhFunction === 'function' && typeof this.fhFunctionObject === 'object') {

      this.fhFunction(this.fhFunctionObject, function (res) {

        var res = res || {};

        me.fireEvent('conditionTrue', res);

      }, function (res) {

        var res = res || {};

        me.fireEvent('conditionFalse', res);

      });

    }
  }

});




/**
 * Manages the set of {@link Ext.ux.Condition} objects that return async callbacks/errbacks. Depending on the configuration, can call each {@link Ext.ux.Condition#fnFunction} in parallel, or can wait until previous condition executes callback/errback. 
 * It also checks the conditions config and prepares the correct callback/errback. 
 * 
 */

Ext.ux.AsyncActionQueue = Ext.extend(Ext.util.Observable, {

  constructor: function (config) {

    this.conditions = [];

    this.actions = {};

    Ext.ux.AsyncActionQueue.superclass.constructor.call(this, config);
    /**
     * @event finishedConditions Triggered after all the conditions' have returned either conditionTrue or conditionFalse events.
     */
    /**
     * @event finishedActions Triggered after all the conditions' callbacks/errbacks have been called.
     */    

    this.addEvents('finishedConditions');

    this.addEvents('finishedActions');

  },

  /**
   * @cfg conditions Queue of {@link Ext.ux.Condition} conditions.
   */   
  conditions: [],
  /**
   * Prepared callback/errbacks ready to be invoked.
   */
  actions: {},
  /**
   * Index of the next condition to run.
   */
  conditionIndex: 0,
  /**
   * Index of the next callback/errback to run.
   */
  actionIndex: 0,
  /**
   * @cfg parallel Determines if the condition checks should be run in parallel or in sync.
   */ 
  parallel: true,

/**
 * Inserts the {@link Ext.ux.Condition} condition into the {@link Ext.ux.AsyncActionQueue#conditions} queue.
 */
  addCondition: function (condition) {

    var me = this;

    condition.on('conditionTrue', function (res) {

      console.log('Time of catching true event of condition "' + condition.id + '": ' + new Date());

      me.actions[condition.id] = me.prepareCallback(condition, res);

      var ready = 0;
      for (var act in me.actions) {
        ready++;
      }

      //if conditions callbabacks were prepared in parallel
      if (ready == me.conditions.length && me.parallel) {

        me.runNextAction();

      }
      //if conditions are not parallel, run action only if it is the first to run
      else if (!me.parallel) {

        me.runNextAction();

      }



    });

    condition.on('conditionFalse', function (res) {


      me.actions[condition.id] = me.prepareErrback(condition, res);

      var ready = 0;
      for (var act in me.actions) {
        ready++;
      }
      if (ready == me.conditions.length && me.parallel) {

        me.runNextAction();

      }
      //if conditions are not parallel, run action only if it is the first to run
      else if (!me.parallel) {

        me.runNextAction();

      }


    });

    this.conditions.push(condition);

  },

  /**
   * Executes the {@link Ext.ux.Condition#fhFunction} of all conditions is parallel.
   */  
  runConditions: function () {

    //if conditions not in parallel run just next condition
    if (!this.parallel) {
      this.runNextCondition();
      return;
    }

    var me = this;

    for (var c = 0; c < this.conditions.length; c++) {

      var condition = this.conditions[c];

      condition.runCondition();

    }

    if (c >= this.conditions.length) {
      this.onConditionsFinished();
      return;
    }

  },

  /**
   * Executes the {@link Ext.ux.Condition#fhFunction} of the condition at index {@link Ext.ux.AsyncActionQueue#conditionIndex}
   */   
  runNextCondition: function () {

    if (this.conditionIndex >= this.conditions.length) {
      this.onConditionsFinished();
      return;
    }

    var nextCondition = this.conditions[this.conditionIndex];


    this.conditionIndex++;

    nextCondition.runCondition();

  },
  /**
   * Executes the {@link Ext.ux.Condition} condition callback/errback at index {@link Ext.ux.AsyncActionQueue#actionIndex}
   */ 
  runNextAction: function () {

    if (this.actionIndex >= this.conditions.length) {
      return;
    }

    var nextActionName = this.conditions[this.actionIndex].id;

    var action = this.actions[nextActionName];

    this.actionIndex++;

    if (typeof action === 'function') {

      action();

    }

  },
  /**
   * Private function that prepares the {@link Ext.ux.Condition} condition callback and that will sit in the {@link Ext.ux.AsyncActionQueue#actions} object
   */ 
  prepareCallback: function (condition, res) {

    var me = this;

    return function () {

      if (typeof condition.trueCallback === 'function') {
        condition.trueCallback(res);
        if (me.parallel) {
          me.runNextAction();
        }
        else if (!me.parallel) {
          me.runNextCondition();
        }
      }
      else if (typeof condition.trueMsg === 'object') {
        me.invokeMsg(condition.trueMsg, condition.trueMsgCallback, condition.trueMsgErrback, condition.trueMsgCallbackStop, condition.trueMsgErrbackStop);
      }
      else if (me.parallel) {
        me.runNextAction();
      }
      else if (!me.parallel) {
        me.runNextCondition();
      }

      if (me.actionIndex >= me.conditions.length) {

        me.onActionsFinished();

      }

    };

  },
  /**
   * Private function that prepares the {@link Ext.ux.Condition} condition errback and that will sit in the {@link Ext.ux.AsyncActionQueue#actions} object
   */ 
  prepareErrback: function (condition, res) {

    var me = this;

    return function () {

      if (typeof condition.falseCallback === 'function') {
        condition.falseCallback(res);
        if (me.parallel) {
          me.runNextAction();
        }
        else if (!me.parallel) {
          me.runNextCondition();
        }
      }
      else if (typeof condition.falseMsg === 'object') {
        me.invokeMsg(condition.falseMsg, condition.falseMsgCallback, condition.falseMsgErrback, condition.falseMsgCallbackStop, condition.falseMsgErrbackStop);
      }
      else if (me.parallel) {
        me.runNextAction();
      }
      else if (!me.parallel) {
        me.runNextCondition();
      }


      if (me.actionIndex >= me.conditions.length) {

        me.onActionsFinished();

      }

    };

  },

  /**
   * Fires the {@link Ext.ux.AsyncActionQueue#event-finishedActions} event
   */ 
  onActionsFinished: function () {

    if(!this.finished) {
    
      this.fireEvent('finishedActions');
      this.finished=true;
    }

  },

  /**
   * Fires the {@link Ext.ux.AsyncActionQueue#event-finishedConditions} event
   */ 
  onConditionsFinished: function () {

    this.fireEvent('finishedConditions');

  },
  
  /**
   * Private function that displays the Ext.MessageBox message.
   */ 

  invokeMsg: function (config, callback, errback, callbackStop, errbackStop) {

    var me = this;

    var config = config || {};

    if (typeof config.title === 'function') {
      config.title = config.title();
    }

    if (typeof config.msg === 'function') {
      config.msg = config.msg();
    }

    var configFn = function (buttonId, opt, config) {
      // if user response yes or ok
      if (buttonId == 'yes' || buttonId == 'ok') {
        if (typeof callback === 'function') {
          callback();
        }
        if (callbackStop === false) {
          if (!me.parallel) {
            me.runNextCondition();
          }
          else {
            me.runNextAction();
          }
        }
      }
      // if user response no
      else if (buttonId == 'no' || buttonId == 'cancel') {
        if (typeof errback === 'function') {
          errback();
        }
        if (errbackStop === false) {
          if (!me.parallel) {
            me.runNextCondition();
          }
          else {
            me.runNextAction();
          }
        }
      }

    };

    config.fn = function(buttonId, opt, config){
    		//prevent clicking the button twice+
    		Ext.Msg.setDisabled();
    		//run config.fn in background
            setTimeout(function(){ configFn(buttonId, opt, config);},100);      		
    };    

    Ext.Msg = new Ext.MessageBox({
      showAnimation: {
        type: 'pop',
        duration: 1
      },
      enterAnimation: {
        type: 'pop',
        duration: 1
      },
      exitAnimation: {
        type: 'pop',
        duration: 1
      }
    });

    Ext.Msg.on({
      hide: function (component) {
        component.destroy();
      }
    });

    console.log('Time of message show: ' + new Date());


    Ext.Msg.show(config);
  }



});
