
/**
 * JSON writer
 */
Ext.ux.FHJsonWriter = Ext.extend(Ext.data.JsonWriter, {
    /**
     * Formats the data for each record before sending it to the server. This
     * method should be overridden to format the data in a way that differs from the default.
     * @param {Object} record The record that we are writing to the server.
     * @return {Object} An object literal of name/value keys to be written to the server.
     * By default this method returns the data property on the record.
     */
    write: function(records) {
        var ln = records.length,
            i = 0,
            data = [];

        for (; i < ln; i++) {
            data.push(this.getRecordData(records[i]));
        }
        return data;
    },

    /**
     * returns record raw data with associations.
     */
    getRecordData: function(record) {
        //Setup variables
        var me = this,
            i, association, childStore;
        var data = record.data;
        //data = record.data;
        //Iterate over all the hasMany associations
        for (i = 0; i < record.associations.length; i++) {
            association = record.associations.get(i);
            if (association.type == 'hasMany') {
                data[association.name] = [];
                if(record[association.storeName]){
	                childStore = record[association.storeName];
	                //Iterate over all the children in the current association
	                childStore.each(function(childRecord) {
	                    //Recursively get the record data for children (depth first)
	                    var childData = this.getRecordData.call(this, childRecord);
	                    data[association.name].push(childData);
	                },me);
                }
            }
        }
        return data;
    }
});

Ext.data.WriterMgr.registerType('fhjson', Ext.ux.FHJsonWriter);